package apigateway

import (
	"context"
	"encoding/json"
	"fmt"
	"math/big"
	"net"
	"net/http"
	"strings"

	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/crypto"
	"github.com/ethereum/go-ethereum/log"
	nebulav1 "github.com/exchain/go-exchain/exchain/protocol/gen/go/nebula/v1"
	orderbookv1 "github.com/exchain/go-exchain/exchain/protocol/gen/go/orderbook/v1"
	"github.com/exchain/process/engine"
	"github.com/exchain/process/orderbook"
	"github.com/exchain/process/types"
	"github.com/exchain/process/utils"
	"github.com/grpc-ecosystem/grpc-gateway/runtime"
	"github.com/holiman/uint256"
	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
)

type server struct {
	eg *engine.Engine
	orderbookv1.UnimplementedOrderbookServer
}

func (s *server) PlaceLimitOrder(ctx context.Context, req *orderbookv1.PlaceLimitOrderRequest) (resp *orderbookv1.Response, err error) {
	resp = new(orderbookv1.Response)
	side := orderbook.Buy
	tx := req.Transaction.Tx.(*nebulav1.Transaction_LimitTx)

	if tx.LimitTx.Side == nebulav1.OrderSide_SELL {
		side = orderbook.Sell
	}

	coins := strings.Split(tx.LimitTx.Pair, "-")
	baseCoin, quoteCoin := coins[0], coins[1]

	quantity := new(uint256.Int).SetBytes(tx.LimitTx.Quantity)
	price := new(uint256.Int).SetBytes(tx.LimitTx.Price)
	nonce := new(big.Int).SetBytes(req.Transaction.Nonce)
	orderID := utils.GenOrderID()
	_, res, err := s.eg.ProcessLimitOrder(orderID,
		types.Coin(baseCoin),
		types.Coin(quoteCoin),
		common.HexToAddress(req.Transaction.User),
		side,
		price,
		quantity,
		uint64(nonce.Int64()),
		req.Transaction.Proxy,
		utils.CombineRSV(req.Transaction.Signature.R, req.Transaction.Signature.S, uint8(req.Transaction.Signature.V)))
	if err != nil {
		return
	}

	data, _ := json.Marshal(res)

	return &orderbookv1.Response{Data: data}, nil
}

func (s *server) PlaceMarketOrder(ctx context.Context, req *orderbookv1.PlaceMarketOrderRequest) (resp *orderbookv1.Response, err error) {
	resp = new(orderbookv1.Response)
	side := orderbook.Buy
	tx := req.Transaction.Tx.(*nebulav1.Transaction_MarketTx)

	if tx.MarketTx.Side == nebulav1.OrderSide_SELL {
		side = orderbook.Sell
	}

	coins := strings.Split(tx.MarketTx.Pair, "-")
	baseCoin, quoteCoin := coins[0], coins[1]
	quantity := new(uint256.Int).SetBytes(tx.MarketTx.Quantity)
	nonce := new(big.Int).SetBytes(req.Transaction.Nonce)
	orderID := utils.GenOrderID()

	_, res, err := s.eg.ProcessMarketOrder(orderID,
		types.Coin(baseCoin),
		types.Coin(quoteCoin),
		common.HexToAddress(req.Transaction.User),
		side,
		quantity,
		uint64(nonce.Int64()),
		req.Transaction.Proxy,
		utils.CombineRSV(req.Transaction.Signature.R, req.Transaction.Signature.S, uint8(req.Transaction.Signature.V)))
	if err != nil {
		return
	}

	data, _ := json.Marshal(res)

	return &orderbookv1.Response{Data: data}, nil
}

func (s *server) CancelOrder(ctx context.Context, req *orderbookv1.CancelOrderRequest) (resp *orderbookv1.Response, err error) {
	tx := req.Transaction.Tx.(*nebulav1.Transaction_CancelTx)
	orderID := tx.CancelTx.OrderId
	coins := strings.Split(tx.CancelTx.Pair, "-")
	baseCoin, quoteCoin := coins[0], coins[1]
	nonce := new(big.Int).SetBytes(req.Transaction.Nonce)
	err = s.eg.ProcessCancelOrder(orderID,
		types.Coin(baseCoin),
		types.Coin(quoteCoin),
		common.HexToAddress(req.Transaction.User),
		nonce.Uint64(),
		req.Transaction.Proxy,
		utils.CombineRSV(req.Transaction.Signature.R, req.Transaction.Signature.S, uint8(req.Transaction.Signature.V)))

	return &orderbookv1.Response{}, err
}

func (s *server) SignProxy(ctx context.Context, req *orderbookv1.SignProxyRequest) (resp *orderbookv1.Response, err error) {
	tx := req.Transaction.Tx.(*nebulav1.Transaction_SignProxyTx)

	nonce := new(big.Int).SetBytes(req.Transaction.Nonce)
	pubkey, err := crypto.UnmarshalPubkey(tx.SignProxyTx.SignerProxy)
	if err != nil {
		return
	}
	err = s.eg.PorcessSignProxy(common.HexToAddress(req.Transaction.User),
		crypto.PubkeyToAddress(*pubkey),
		nonce.Uint64(),
		utils.CombineRSV(req.Transaction.Signature.R, req.Transaction.Signature.S, uint8(req.Transaction.Signature.V)))
	return &orderbookv1.Response{}, err
}

func StartServer(grpcPort int, gwPort int, eg *engine.Engine) error {
	defer log.Info("grpc and gateway server stopped")
	log.Info("starting grpc and gateway server", "grpcPort", grpcPort, "gwPort", gwPort)
	ctx := context.Background()
	ctx, cancel := context.WithCancel(ctx)
	defer cancel()

	gServer := &server{
		eg: eg,
	}
	listener, err := net.Listen("tcp", fmt.Sprintf("0.0.0.0:%d", grpcPort))
	if err != nil {
		log.Error("failed to listen", "err", err)
		return err
	}
	errCh := make(chan error)
	grpcServer := grpc.NewServer()
	orderbookv1.RegisterOrderbookServer(grpcServer, gServer)
	go func(ch chan error) {
		var e = grpcServer.Serve(listener)
		ch <- e
	}(errCh)

	mux := runtime.NewServeMux()
	opts := []grpc.DialOption{grpc.WithTransportCredentials(insecure.NewCredentials())}
	err = orderbookv1.RegisterOrderbookHandlerFromEndpoint(ctx, mux, fmt.Sprintf("127.0.0.1:%d", grpcPort), opts)
	if err != nil {
		return err
	}
	go func(ch chan error) {
		var e = http.ListenAndServe(fmt.Sprintf("0.0.0.0:%d", gwPort), mux)
		ch <- e
	}(errCh)

	e := <-errCh
	return e
}
